import { groupBy } from "lodash";
import { Endianness } from "../../Endianness";
import { Cursor } from "../../cursor/Cursor";
import { ResizableBufferCursor } from "../../cursor/ResizableBufferCursor";
import { ResizableBuffer } from "../../ResizableBuffer";
import { Vec3 } from "../../vector";
import { WritableCursor } from "../../cursor/WritableCursor";
import { assert } from "../../../util";
import { LogManager } from "../../../Logger";

const logger = LogManager.get("core/data_formats/parsing/quest/dat");

const OBJECT_SIZE = 68;
const NPC_SIZE = 72;

export type DatFile = {
    readonly objs: readonly DatObject[];
    readonly npcs: readonly DatNpc[];
    readonly events: readonly DatEvent[];
    readonly unknowns: readonly DatUnknown[];
};

export type DatEntity = {
    readonly type_id: number;
    readonly section_id: number;
    readonly position: Vec3;
    readonly rotation: Vec3;
    readonly area_id: number;
    readonly unknown: readonly number[][];
};

export type DatObject = DatEntity & {
    readonly id: number;
    readonly group_id: number;
    readonly properties: readonly number[];
};

export type DatNpc = DatEntity & {
    readonly wave: number;
    readonly wave2: number;
    readonly scale: Vec3;
    readonly npc_id: number;
    readonly script_label: number;
    readonly roaming: number;
};

export type DatEvent = {
    readonly id: number;
    readonly section_id: number;
    readonly wave: number;
    readonly delay: number;
    readonly actions: readonly DatEventAction[];
    readonly area_id: number;
    readonly unknown: number;
};

export enum DatEventActionType {
    SpawnNpcs = 0x8,
    Unlock = 0xa,
    Lock = 0xb,
    TriggerEvent = 0xc,
}

export type DatEventAction =
    | DatEventActionSpawnNpcs
    | DatEventActionUnlock
    | DatEventActionLock
    | DatEventActionTriggerEvent;

export type DatEventActionSpawnNpcs = {
    readonly type: DatEventActionType.SpawnNpcs;
    readonly section_id: number;
    readonly appear_flag: number;
};

export type DatEventActionUnlock = {
    readonly type: DatEventActionType.Unlock;
    readonly door_id: number;
};

export type DatEventActionLock = {
    readonly type: DatEventActionType.Lock;
    readonly door_id: number;
};

export type DatEventActionTriggerEvent = {
    readonly type: DatEventActionType.TriggerEvent;
    readonly event_id: number;
};

export type DatUnknown = {
    readonly entity_type: number;
    readonly total_size: number;
    readonly area_id: number;
    readonly entities_size: number;
    readonly data: number[];
};

export function parse_dat(cursor: Cursor): DatFile {
    const objs: DatObject[] = [];
    const npcs: DatNpc[] = [];
    const events: DatEvent[] = [];
    const unknowns: DatUnknown[] = [];

    while (cursor.bytes_left) {
        const entity_type = cursor.u32();
        const total_size = cursor.u32();
        const area_id = cursor.u32();
        const entities_size = cursor.u32();

        if (entity_type === 0) {
            break;
        } else {
            if (entities_size !== total_size - 16) {
                throw Error(
                    `Malformed DAT file. Expected an entities size of ${
                        total_size - 16
                    }, got ${entities_size}.`,
                );
            }

            const entities_cursor = cursor.take(entities_size);

            if (entity_type === 1) {
                parse_objects(entities_cursor, area_id, objs);
            } else if (entity_type === 2) {
                parse_npcs(entities_cursor, area_id, npcs);
            } else if (entity_type === 3) {
                parse_events(entities_cursor, area_id, events);
            } else {
                // Unknown entity types 4 and 5 (challenge mode).
                unknowns.push({
                    entity_type,
                    total_size,
                    area_id,
                    entities_size,
                    data: cursor.u8_array(entities_size),
                });
            }

            if (entities_cursor.bytes_left) {
                logger.warn(
                    `Read ${entities_cursor.position} bytes instead of expected ${entities_cursor.size} for entity type ${entity_type}.`,
                );
            }
        }
    }

    return { objs, npcs, events, unknowns };
}

export function write_dat({ objs, npcs, events, unknowns }: DatFile): ResizableBuffer {
    const buffer = new ResizableBuffer(
        objs.length * (16 + OBJECT_SIZE) +
            npcs.length * (16 + NPC_SIZE) +
            unknowns.reduce((a, b) => a + b.total_size, 0),
    );
    const cursor = new ResizableBufferCursor(buffer, Endianness.Little);

    write_objects(cursor, objs);

    write_npcs(cursor, npcs);

    write_events(cursor, events);

    for (const unknown of unknowns) {
        cursor.write_u32(unknown.entity_type);
        cursor.write_u32(unknown.total_size);
        cursor.write_u32(unknown.area_id);
        cursor.write_u32(unknown.entities_size);
        cursor.write_u8_array(unknown.data);
    }

    // Final header.
    cursor.write_u32(0);
    cursor.write_u32(0);
    cursor.write_u32(0);
    cursor.write_u32(0);

    return buffer;
}

function parse_objects(cursor: Cursor, area_id: number, objs: DatObject[]): void {
    const object_count = Math.floor(cursor.size / OBJECT_SIZE);

    for (let i = 0; i < object_count; ++i) {
        const type_id = cursor.u16();
        const unknown1 = cursor.u8_array(6);
        const id = cursor.u16();
        const group_id = cursor.u16();
        const section_id = cursor.u16();
        const unknown2 = cursor.u8_array(2);
        const position = cursor.vec3_f32();
        const rotation = {
            x: (cursor.i32() / 0xffff) * 2 * Math.PI,
            y: (cursor.i32() / 0xffff) * 2 * Math.PI,
            z: (cursor.i32() / 0xffff) * 2 * Math.PI,
        };
        const properties = [
            cursor.f32(),
            cursor.f32(),
            cursor.f32(),
            cursor.u32(),
            cursor.u32(),
            cursor.u32(),
            cursor.u32(),
        ];

        objs.push({
            type_id,
            id,
            group_id,
            section_id,
            position,
            rotation,
            properties,
            area_id,
            unknown: [unknown1, unknown2],
        });
    }
}

function parse_npcs(cursor: Cursor, area_id: number, npcs: DatNpc[]): void {
    const npc_count = Math.floor(cursor.size / NPC_SIZE);

    for (let i = 0; i < npc_count; ++i) {
        const type_id = cursor.u16();
        const unknown1 = cursor.u8_array(10);
        const section_id = cursor.u16();
        const wave = cursor.u16();
        const wave2 = cursor.u32();
        const position = cursor.vec3_f32();
        const rotation_x = (cursor.i32() / 0xffff) * 2 * Math.PI;
        const rotation_y = (cursor.i32() / 0xffff) * 2 * Math.PI;
        const rotation_z = (cursor.i32() / 0xffff) * 2 * Math.PI;
        const scale = cursor.vec3_f32();
        const npc_id = cursor.f32();
        const script_label = cursor.f32();
        const roaming = cursor.u32();
        const unknown2 = cursor.u8_array(4);

        npcs.push({
            type_id,
            section_id,
            wave,
            wave2,
            position,
            rotation: { x: rotation_x, y: rotation_y, z: rotation_z },
            scale,
            npc_id,
            script_label,
            roaming,
            area_id,
            unknown: [unknown1, unknown2],
        });
    }
}

function parse_events(cursor: Cursor, area_id: number, events: DatEvent[]): void {
    const actions_offset = cursor.u32();
    cursor.seek(4); // Always 0x10
    const event_count = cursor.u32();
    cursor.seek(3); // Always 0
    const event_type = cursor.u8();

    if (event_type === 0x32) {
        throw new Error("Can't parse challenge mode quests yet.");
    }

    cursor.seek_start(actions_offset);
    const actions_cursor = cursor.take(cursor.bytes_left);
    cursor.seek_start(16);

    for (let i = 0; i < event_count; ++i) {
        const id = cursor.u32();
        cursor.seek(4); // Always 0x100
        const section_id = cursor.u16();
        const wave = cursor.u16();
        const delay = cursor.u16();
        const unknown = cursor.u16(); // "wavesetting"?
        const event_actions_offset = cursor.u32();

        let actions: DatEventAction[] = [];

        if (event_actions_offset < actions_cursor.size) {
            actions_cursor.seek_start(event_actions_offset);
            actions = parse_event_actions(actions_cursor);
        } else {
            logger.warn(`Invalid event actions offset ${event_actions_offset} for event ${id}.`);
        }

        events.push({
            id,
            section_id,
            wave,
            delay,
            actions,
            area_id,
            unknown,
        });
    }

    if (cursor.position !== actions_offset) {
        logger.warn(
            `Read ${cursor.position - 16} bytes of event data instead of expected ${
                actions_offset - 16
            }.`,
        );
    }

    let last_u8 = 0xff;

    while (actions_cursor.bytes_left) {
        last_u8 = actions_cursor.u8();

        if (last_u8 !== 0xff) {
            break;
        }
    }

    if (last_u8 !== 0xff) {
        actions_cursor.seek(-1);
    }

    // Make sure the cursor position represents the amount of bytes we've consumed.
    cursor.seek_start(actions_offset + actions_cursor.position);
}

function parse_event_actions(cursor: Cursor): DatEventAction[] {
    const actions: DatEventAction[] = [];

    outer: while (cursor.bytes_left) {
        const type = cursor.u8();

        switch (type) {
            case 1:
                break outer;

            case DatEventActionType.SpawnNpcs:
                actions.push({
                    type: DatEventActionType.SpawnNpcs,
                    section_id: cursor.u16(),
                    appear_flag: cursor.u16(),
                });
                break;

            case DatEventActionType.Unlock:
                actions.push({
                    type: DatEventActionType.Unlock,
                    door_id: cursor.u16(),
                });
                break;

            case DatEventActionType.Lock:
                actions.push({
                    type: DatEventActionType.Lock,
                    door_id: cursor.u16(),
                });
                break;

            case DatEventActionType.TriggerEvent:
                actions.push({
                    type: DatEventActionType.TriggerEvent,
                    event_id: cursor.u32(),
                });
                break;

            default:
                logger.warn(`Unexpected event action type ${type}.`);
                break outer;
        }
    }

    return actions;
}

function write_objects(cursor: WritableCursor, objs: readonly DatObject[]): void {
    const grouped_objs = groupBy(objs, obj => obj.area_id);
    const obj_area_ids = Object.keys(grouped_objs)
        .map(key => parseInt(key, 10))
        .sort((a, b) => a - b);

    for (const area_id of obj_area_ids) {
        const area_objs = grouped_objs[area_id];
        const entities_size = area_objs.length * OBJECT_SIZE;
        cursor.write_u32(1); // Entity type
        cursor.write_u32(entities_size + 16);
        cursor.write_u32(area_id);
        cursor.write_u32(entities_size);

        for (const obj of area_objs) {
            assert(
                obj.unknown.length === 2,
                () => `unknown should be of length 2, was ${obj.unknown.length}`,
            );

            cursor.write_u16(obj.type_id);

            assert(
                obj.unknown[0].length === 6,
                () => `unknown[0] should be of length 6, was ${obj.unknown[0].length}`,
            );

            cursor.write_u8_array(obj.unknown[0]);
            cursor.write_u16(obj.id);
            cursor.write_u16(obj.group_id);
            cursor.write_u16(obj.section_id);

            assert(
                obj.unknown[1].length === 2,
                () => `unknown[1] should be of length 2, was ${obj.unknown[1].length}`,
            );

            cursor.write_u8_array(obj.unknown[1]);
            cursor.write_vec3_f32(obj.position);
            cursor.write_i32(Math.round((obj.rotation.x / (2 * Math.PI)) * 0xffff));
            cursor.write_i32(Math.round((obj.rotation.y / (2 * Math.PI)) * 0xffff));
            cursor.write_i32(Math.round((obj.rotation.z / (2 * Math.PI)) * 0xffff));

            assert(
                obj.properties.length === 7,
                () => `properties should be of length 7, was ${obj.properties.length}`,
            );

            cursor.write_f32(obj.properties[0]);
            cursor.write_f32(obj.properties[1]);
            cursor.write_f32(obj.properties[2]);
            cursor.write_u32(obj.properties[3]);
            cursor.write_u32(obj.properties[4]);
            cursor.write_u32(obj.properties[5]);
            cursor.write_u32(obj.properties[6]);
        }
    }
}

function write_npcs(cursor: WritableCursor, npcs: readonly DatNpc[]): void {
    const grouped_npcs = groupBy(npcs, npc => npc.area_id);
    const npc_area_ids = Object.keys(grouped_npcs)
        .map(key => parseInt(key, 10))
        .sort((a, b) => a - b);

    for (const area_id of npc_area_ids) {
        const area_npcs = grouped_npcs[area_id];
        const entities_size = area_npcs.length * NPC_SIZE;
        cursor.write_u32(2); // Entity type
        cursor.write_u32(entities_size + 16);
        cursor.write_u32(area_id);
        cursor.write_u32(entities_size);

        for (const npc of area_npcs) {
            assert(
                npc.unknown.length === 2,
                () => `unknown should be of length 2, was ${npc.unknown.length}`,
            );

            cursor.write_u16(npc.type_id);

            assert(
                npc.unknown[0].length === 10,
                () => `unknown[0] should be of length 10, was ${npc.unknown[0].length}`,
            );

            cursor.write_u8_array(npc.unknown[0]);
            cursor.write_u16(npc.section_id);
            cursor.write_u16(npc.wave);
            cursor.write_u32(npc.wave2);
            cursor.write_vec3_f32(npc.position);
            cursor.write_i32(Math.round((npc.rotation.x / (2 * Math.PI)) * 0xffff));
            cursor.write_i32(Math.round((npc.rotation.y / (2 * Math.PI)) * 0xffff));
            cursor.write_i32(Math.round((npc.rotation.z / (2 * Math.PI)) * 0xffff));
            cursor.write_vec3_f32(npc.scale);
            cursor.write_f32(npc.npc_id);
            cursor.write_f32(npc.script_label);
            cursor.write_u32(npc.roaming);

            assert(
                npc.unknown[1].length === 4,
                () => `unknown[1] should be of length 4, was ${npc.unknown[1].length}`,
            );

            cursor.write_u8_array(npc.unknown[1]);
        }
    }
}

function write_events(cursor: WritableCursor, events: readonly DatEvent[]): void {
    const grouped_events = groupBy(events, event => event.area_id);
    const event_area_ids = Object.keys(grouped_events)
        .map(key => parseInt(key, 10))
        .sort((a, b) => a - b);

    for (const area_id of event_area_ids) {
        const area_events = grouped_events[area_id];

        // Standard header.
        cursor.write_u32(3); // Entity type
        const total_size_offset = cursor.position;
        cursor.write_u32(0); // Placeholder for the total size.
        cursor.write_u32(area_id);
        const entities_size_offset = cursor.position;
        cursor.write_u32(0); // Placeholder for the entities size.

        // Event header.
        const start_pos = cursor.position;
        // TODO: actual event size is dependent on the event type (challenge mode).
        // Absolute offset.
        const actions_offset = start_pos + 16 + 20 * area_events.length;
        cursor.size = Math.max(actions_offset, cursor.size);

        cursor.write_u32(actions_offset - start_pos);
        cursor.write_u32(0x10);
        cursor.write_u32(area_events.length);
        cursor.write_u32(0); // TODO: write event type (challenge mode).

        // Relative offset.
        let event_actions_offset = 0;

        for (const event of area_events) {
            cursor.write_u32(event.id);
            cursor.write_u32(0x10000);
            cursor.write_u16(event.section_id);
            cursor.write_u16(event.wave);
            cursor.write_u16(event.delay);
            cursor.write_u16(event.unknown);
            cursor.write_u32(event_actions_offset);
            const next_event_pos = cursor.position;

            cursor.seek_start(actions_offset + event_actions_offset);

            for (const action of event.actions) {
                cursor.write_u8(action.type);

                switch (action.type) {
                    case DatEventActionType.SpawnNpcs:
                        cursor.write_u16(action.section_id);
                        cursor.write_u16(action.appear_flag);
                        break;

                    case DatEventActionType.Unlock:
                        cursor.write_u16(action.door_id);
                        break;

                    case DatEventActionType.Lock:
                        cursor.write_u16(action.door_id);
                        break;

                    case DatEventActionType.TriggerEvent:
                        cursor.write_u32(action.event_id);
                        break;

                    default:
                        // Need to cast because TypeScript infers action to be `never`.
                        throw new Error(`Unknown event action type ${(action as any).type}.`);
                }
            }

            // End of event actions.
            cursor.write_u8(1);

            event_actions_offset = cursor.position - actions_offset;

            cursor.seek_start(next_event_pos);
        }

        cursor.seek_start(actions_offset + event_actions_offset);

        while ((cursor.position - actions_offset) % 4 !== 0) {
            cursor.write_u8(0xff);
        }

        const end_pos = cursor.position;

        cursor.seek_start(total_size_offset);
        cursor.write_u32(end_pos - start_pos + 16);

        cursor.seek_start(entities_size_offset);
        cursor.write_u32(end_pos - start_pos);

        cursor.seek_start(end_pos);
    }
}
